/*
* Copyright (c) Gengtu Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <linux/printk.h>
#if defined(__linux__)
 #include <linux/version.h>
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0))
  #include <linux/stdarg.h>
 #else
  #include <stdarg.h>
 #endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0) */
#else
 #include <stdarg.h>
#endif /* __linux__ */
#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>

#include "gt_debug.h"
#include "gt_misc.h"
#include "gt_mm.h"
#include "gt_timer.h"
#include "gt_debugfs.h"
#include "gt_lock.h"

unsigned int s_dpu_debug = 0x00;
MODULE_PARM_DESC(s_dpu_debug, "Enable debug output, where each bit enables a debug category.\n"
		"\t\tBit 0 (0x01) (drm & fbdev code)\n"
		"\t\tBit 1 (0x02) (modeset code)\n"
		"\t\tBit 2 (0x04) (gem code)\n"
		"\t\tBit 3 (0x08) (dpu code)\n"
		"\t\tBit 4 (0x10) (hdmi code)\n"
		"\t\tBit 5 (0x20) (dp code)\n"
		"\t\tBit 6 (0x40) (lvds code)\n"
		"\t\tBit 7 (0x80) (vga code)\n");
module_param_named(s_dpu_debug, s_dpu_debug, uint, 0600);

bool s_dpu_open_log = false;
MODULE_PARM_DESC(s_dpu_open_log, "Enable save output");
module_param_named(s_dpu_open_log, s_dpu_open_log, bool, 0600);

#define DPU_LOG_ITEM_LEN 150
#define DPU_LOG_PRINT_LEN (DPU_LOG_ITEM_LEN + 34 + 1)

typedef enum log_mod_e {
	GTDPU_LOG_MOD_FATAL = 0,
	GTDPU_LOG_MOD_CRTC,
	GTDPU_LOG_MOD_HDMI,
	GTDPU_LOG_MOD_DP,
	GTDPU_LOG_MOD_VGA,
	GTDPU_LOG_MOD_LVDS,
	GTDPU_LOG_MOD_DRM,
	GTDPU_LOG_MOD_KMS,
	GTDPU_LOG_MOD_GEM,
	GTDPU_LOG_MOD_MAX,
} log_mod;

typedef struct node {
	char   *data;
	struct node *next;
} circular_node;

typedef struct {
	circular_node *head;
	circular_node *tail;
	unsigned long read_size;
	char *read_buf;
	int size;
	int max_size;
} circular_list;

struct gtdpu_log {
	circular_list *kfifo[GTDPU_LOG_MOD_MAX];
	struct mutex *log_mutex;
	loff_t pos;
	/* record current target want get */
	log_mod mod;
};

static struct gtdpu_log *gt_log = NULL;
static char *s_dpu_log_mod_name[GTDPU_LOG_MOD_MAX] = {
	"[FATAL]", "[CRTC]", "[HDMI]", "[DP]", "[VGA]", "[LVDS]", "[DRM]", "[KMS]", "[GEM]"
};

static void gtdpu_log_info(gt_dev *dev, log_mod mod, const char *format, ...);

void gt_dev_printk(char *debug_level, gt_dev *dev, const char *format, ...)
{
	struct va_format vaf;
	va_list args;

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	dev_printk(debug_level, dev, "%pV", &vaf);

	va_end(args);
}

void gt_printk(const char *format, ...)
{
	struct va_format vaf;
	va_list args;

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	printk("%pV", &vaf);

	va_end(args);
}

void gt_printk_ratelimited(const char *format, ...)
{
	struct va_format vaf;
	va_list args;

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	printk_ratelimited("%pV", &vaf);

	va_end(args);
}

static log_mod gtdpu_get_mod_by_category(unsigned int category)
{
	log_mod mod = GTDPU_LOG_MOD_FATAL;

	if (category & DPU_UT_HDMI) {
		mod = GTDPU_LOG_MOD_HDMI;
	} else if (category & DPU_UT_DPU) {
		mod = GTDPU_LOG_MOD_CRTC;
	} else if (category & DPU_UT_DP) {
		mod = GTDPU_LOG_MOD_DP;
	} else if (category & DPU_UT_LVDS) {
		mod = GTDPU_LOG_MOD_LVDS;
	} else if (category & DPU_UT_VGA) {
		mod = GTDPU_LOG_MOD_VGA;
	} else if (category & DPU_UT_MEM) {
		mod = GTDPU_LOG_MOD_GEM;
	} else if (category & DPU_UT_DRM) {
		mod = GTDPU_LOG_MOD_DRM;
	} else if (category & DPU_UT_KMS) {
		mod = GTDPU_LOG_MOD_KMS;
	} else {
		mod = GTDPU_LOG_MOD_FATAL;
	}

	return mod;
}

void gtdpu_info(gt_dev *dev, unsigned int category, const char *format, ...)
{
	log_mod mod = GTDPU_LOG_MOD_FATAL;
	struct va_format vaf;
	va_list args;

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	if (category != DPU_UT_CURSOR && category != DPU_UT_GAMMA) {
		mod = gtdpu_get_mod_by_category(category);
		gtdpu_log_info(dev, mod, "[drm:[info] %ps] %pV", __builtin_return_address(0), &vaf);
	}

#if !defined(NO_HARDWARE)
	if (!(s_dpu_debug & category)) {
		va_end(args);
		return;
	}
#endif

	if (dev) {
		gt_dev_printk(KERN_INFO, dev, "[drm:[info] %ps] %pV", __builtin_return_address(0), &vaf);
	} else {
		gt_printk(KERN_INFO "[drm:[info] %ps] %pV", __builtin_return_address(0), &vaf);
	}

	va_end(args);
}

void gt_print_hex_dump(const char *level, const char *prefix_str,
				  int prefix_type, int rowsize, int groupsize,
				  const void *buf, size_t len, bool ascii)
{
	log_mod mod = GTDPU_LOG_MOD_KMS;
	const unsigned char *ptr = buf;
	int i, linelen, remaining = len;
	unsigned char linebuf[32 * 3 + 2 + 32 + 1];

	if (rowsize != 16 && rowsize != 32) {
		rowsize = 16;
	}

	for (i = 0; i < len; i += rowsize) {
		linelen = min(remaining, rowsize);
		remaining -= rowsize;

		hex_dump_to_buffer(ptr + i, linelen, rowsize, groupsize,
				   linebuf, sizeof(linebuf), ascii);

		gtdpu_log_info(NULL, mod, "%s%s\n", prefix_str, linebuf);
	}
}

void gtdpu_warn(gt_dev *dev, const char *format, ...)
{
	log_mod mod = GTDPU_LOG_MOD_FATAL;
	struct va_format vaf;
	va_list args;

	/* fixme: warn close printk */
	if (!s_dpu_open_log) {
		return ;
	}

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	if (dev) {
		gt_dev_printk(KERN_WARNING, dev, "[drm:[warn] %ps] %pV", __builtin_return_address(0), &vaf);
	} else {
		gt_printk(KERN_WARNING "[drm:[warn] %ps] %pV", __builtin_return_address(0), &vaf);
	}

	gtdpu_log_info(dev, mod, "[drm:[warn] %ps] %pV", __builtin_return_address(0), &vaf);

	va_end(args);
}

void gtdpu_err(gt_dev *dev, const char *format, ...)
{
	log_mod mod = GTDPU_LOG_MOD_FATAL;
	struct va_format vaf;
	va_list args;

	va_start(args, format);
	vaf.fmt = format;
	vaf.va = &args;

	if (dev) {
		gt_dev_printk(KERN_WARNING, dev, "[drm:[ err] %ps] %pV", __builtin_return_address(0), &vaf);
	} else {
		gt_printk(KERN_WARNING "[drm:[ err] %ps] %pV", __builtin_return_address(0), &vaf);
	}

	gtdpu_log_info(dev, mod, "[drm:[ err] %ps] %pV", __builtin_return_address(0), &vaf);

	va_end(args);
}

void gt_seq_printf(gt_seq_file *m, const char *f, ...)
{
	va_list args;

	va_start(args, f);
	gt_seq_vprintf(m, f, args);
	va_end(args);
}

static circular_list *init_circular_list(int max_size)
{
	circular_list *log_handle = (circular_list *)gt_kzalloc_kernel(sizeof(circular_list));

	if (!log_handle) {
		return NULL;
	}

	log_handle->head = NULL;
	log_handle->tail = NULL;
	log_handle->size = 0;
	log_handle->max_size = max_size;

	return log_handle;
}

static void insert_node(circular_list *list_handle, char *data)
{
	circular_node *new_node  = NULL;
	circular_node *temp_node = NULL;

	if (!list_handle) {
		return;
	}

	if (list_handle->size < list_handle->max_size) {
		/* insert tail node */
		new_node = (circular_node *)gt_kzalloc_kernel(sizeof(circular_node));
		if (!new_node) {
			return;
		}
		/* copy printk data */
		new_node->data = kstrndup(data, DPU_LOG_PRINT_LEN, GFP_KERNEL);
		if (!new_node->data) {
			if (new_node) {
				gt_kfree(new_node);
				new_node = NULL;
			}
			return;
		}

		if (list_handle->head == NULL) {
			list_handle->head = new_node;
		} else {
			list_handle->tail->next = new_node;
		}

		list_handle->tail = new_node;
		list_handle->tail->next = list_handle->head;
		list_handle->size++;
	} else {
		/* remove header node, and insert_node */
		temp_node = list_handle->head;
		list_handle->head = list_handle->head->next;
		list_handle->size--;
		if (temp_node->data) {
			gt_kfree(temp_node->data);
			temp_node->data = NULL;
		}
		if (temp_node) {
			gt_kfree(temp_node);
			temp_node = NULL;
		}
		list_handle->tail->next = NULL;
		insert_node(list_handle, data);
	}
}

static void clear_circular_list(circular_list *list_handle)
{
	circular_node *current_node = list_handle->head;
	circular_node *temp_node = NULL;
	int i;

	if (!list_handle) {
		return;
	}

	for (i = 0; i < list_handle->size; i++) {
		temp_node = current_node;
		current_node = current_node->next;
		if (temp_node->data) {
			gt_kfree(temp_node->data);
			temp_node->data = NULL;
		}
		if (temp_node) {
			gt_kfree(temp_node);
			temp_node = NULL;
		}
	}

	list_handle->head = NULL;
	list_handle->size = 0;
}

static void free_circular_list(circular_list *list_handle)
{
	circular_node *current_node = list_handle->head;
	circular_node *temp_node = NULL;
	int i;

	if (!list_handle) {
		return;
	}

	for (i = 0; i < list_handle->size; i++) {
		temp_node = current_node;
		current_node = current_node->next;
		if (temp_node->data) {
			gt_kfree(temp_node->data);
			temp_node->data = NULL;
		}
		if (temp_node) {
			gt_kfree(temp_node);
			temp_node = NULL;
		}
	}

	if (list_handle) {
		gt_kfree(list_handle);
		list_handle = NULL;
	}
}

static void copy_circular_list(circular_list *list_handle, char *buf, unsigned int size)
{
	circular_node *current_node = list_handle->head;
	int i;

	for (i = 0; i < size; i++) {
		gt_strlcpy(buf + i * DPU_LOG_PRINT_LEN, current_node->data, DPU_LOG_PRINT_LEN);
		current_node = current_node->next;
	}
}

static bool gtdpu_log_mod_check(log_mod mod)
{
	return ((mod >= GTDPU_LOG_MOD_FATAL) && (mod < GTDPU_LOG_MOD_MAX));
}

static char *get_mod_name_by_id(log_mod mod)
{
	static char *invalid_mod = "invalid_mod";

	if (gtdpu_log_mod_check(mod)) {
		return s_dpu_log_mod_name[mod];
	}

	return invalid_mod;
}

static void gtdpu_log_print(gt_dev *dev, log_mod mod, char *message)
{
	char temp_buffer[DPU_LOG_PRINT_LEN] = {"\0"};
	gt_ktime current_time = gt_ktime_get();
	circular_list *kfifo = NULL;
	unsigned long utime;
	int len = 0, i;

	/* fixme: close kfifo save log */
	if (!s_dpu_open_log) {
		return ;
	}

	if (unlikely(!gt_log)) {
		return;
	}

	gt_mutex_lock(gt_log->log_mutex);

	utime = gt_ktime_to_us(current_time);
	gt_snprintf(temp_buffer, DPU_LOG_ITEM_LEN + 34, "%7.7s:%8.7s(%5.5d) [%5ld.%3.3ld] %s",
			 get_mod_name_by_id(mod), current->comm, current->tgid,
			 utime / 1000 / 1000, utime / 1000 % 1000, message);

	len = gt_strlen(temp_buffer);
	if (len > DPU_LOG_PRINT_LEN) {
		len = DPU_LOG_PRINT_LEN;
	}
	if (len > 0 && temp_buffer[len - 1] != '\n') {
		gt_strlcat(temp_buffer, "\n", sizeof(temp_buffer));
	}

	if (mod == GTDPU_LOG_MOD_KMS) {
		for (i = GTDPU_LOG_MOD_CRTC; i <= GTDPU_LOG_MOD_LVDS; i++) {
			kfifo = gt_log->kfifo[i];
			insert_node(kfifo, temp_buffer);
		}
	}

	kfifo = gt_log->kfifo[mod];
	insert_node(kfifo, temp_buffer);

	gt_mutex_unlock(gt_log->log_mutex);
}

static void gtdpu_log_info(gt_dev *dev, log_mod mod, const char *format, ...)
{
	char temp_buffer[DPU_LOG_ITEM_LEN] = {"\0"};
	struct va_format vaf;
	va_list args;

	/* check debug mod here */
	if (!gtdpu_log_mod_check(mod)) {
		return;
	}

	va_start(args, format);
	vaf.fmt = format;
	vaf.va  = &args;
	gt_vsnprintf(temp_buffer, DPU_LOG_ITEM_LEN, format, args);
	va_end(args);

	gtdpu_log_print(dev, mod, temp_buffer);
}

int gtdpu_log_init(void)
{
	int ret = 0, mod;

	gt_log = gt_kzalloc_kernel(sizeof(*gt_log));
	if (!gt_log) {
		return -ENOMEM;
	}

	gt_log->mod = GTDPU_LOG_MOD_FATAL;

	gt_log->log_mutex = gt_mutex_alloc();

	for (mod = 0; mod < GTDPU_LOG_MOD_MAX; mod++) {
		gt_log->kfifo[mod] = init_circular_list(1000);
	}

	return ret;
}

void gtdpu_log_fini(void)
{
	int i = 0;

	if (unlikely(!gt_log)) {
		return;
	}

	if (gt_log->log_mutex) {
		gt_mutex_free(gt_log->log_mutex);
	}

	for (i = 0; i < GTDPU_LOG_MOD_MAX; i++) {
		free_circular_list(gt_log->kfifo[i]);
	}

	gt_kfree(gt_log);
	gt_log = NULL;
}

/* debufs node this is the entry to userspace */
static int gtdpu_log_open(struct inode *inode, struct file *file)
{
	file->private_data = inode->i_private;

	return 0;
}

static ssize_t gtdpu_log_read(struct file *file,
		char __user * user_buffer, size_t size, loff_t * ppos)
{
	circular_list *kfifo = NULL;
	unsigned long buf_size = 0;
	loff_t position = *ppos;
	size_t ret = 0;

	if (unlikely(!gt_log)) {
		return 0;
	}

	if (!gtdpu_log_mod_check(gt_log->mod)) {
		return 0;
	}

	gt_mutex_lock(gt_log->log_mutex);

	kfifo = gt_log->kfifo[gt_log->mod];
	if (!kfifo) {
		gt_mutex_unlock(gt_log->log_mutex);
		return 0;
	}

	/* only printk DPU_LOG_PRINT_LEN once */
	if (size > DPU_LOG_PRINT_LEN) {
		size = DPU_LOG_PRINT_LEN;
	}

	if (position == 0) {
		/* printk begin
		 * 1. get fifo size
		 * 2. copy fifo data */
		buf_size = kfifo->size * DPU_LOG_PRINT_LEN;
		kfifo->read_size = buf_size;
		if (!buf_size) {
			gt_mutex_unlock(gt_log->log_mutex);
			return 0;
		}

		kfifo->read_buf  = gt_vmalloc(buf_size);
		if (!kfifo->read_buf) {
			gt_mutex_unlock(gt_log->log_mutex);
			return 0;
		}
		gt_memset(kfifo->read_buf, 0, buf_size);

		copy_circular_list(kfifo, kfifo->read_buf, kfifo->size);
		if (buf_size < size) {
			size = buf_size;
		}
	} else {
		buf_size = kfifo->read_size;

		if ((position + size) >= buf_size) {
			size = buf_size - position;
		}
		if ((buf_size - position) <= 0) {
			size = 0;
		}
	}

	if (size && kfifo->read_buf) {
		gt_copy_to_user(user_buffer, kfifo->read_buf + position, size);
	}

	*ppos = position + size;
	ret = size;

	if (size == 0 && kfifo->read_buf) {
		gt_vfree(kfifo->read_buf);
		kfifo->read_buf = NULL;
		kfifo->read_size = 0;
	}

	gt_mutex_unlock(gt_log->log_mutex);

	return ret;
}

/* this maybe use to clear log fifo, to do */
static ssize_t gtdpu_log_write(struct file *file,
		const char __user * user_buffer, size_t count, loff_t * position)
{
	char temp_buffer[64] = {"\0"};
	int err = -EFAULT;

	if (unlikely(!gt_log)) {
		return err;
	}

	gt_mutex_lock(gt_log->log_mutex);
	count = min(count, ARRAY_SIZE(temp_buffer) - 1);
	err   = gt_copy_from_user(temp_buffer, user_buffer, count);
	if (err) {
		goto unlocked;
	}

	if (!gt_strncmp(temp_buffer, "fatal", sizeof("fatal") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_FATAL;
	} else if (!gt_strncmp(temp_buffer, "crtc", sizeof("crtc") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_CRTC;
	} else if (!gt_strncmp(temp_buffer, "hdmi", sizeof("hdmi") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_HDMI;
	} else if (!gt_strncmp(temp_buffer, "dp", sizeof("dp") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_DP;
	} else if (!gt_strncmp(temp_buffer, "vga", sizeof("vga") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_VGA;
	} else if (!gt_strncmp(temp_buffer, "gem", sizeof("gem") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_GEM;
	} else if (!gt_strncmp(temp_buffer, "lvds", sizeof("lvds") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_LVDS;
	} else if (!gt_strncmp(temp_buffer, "drm", sizeof("drm") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_DRM;
	} else if (!gt_strncmp(temp_buffer, "kms", sizeof("kms") - 1)) {
		gt_log->mod = GTDPU_LOG_MOD_KMS;
	} else if (!gt_strncmp(temp_buffer, "clear", sizeof("clear") - 1)) {
		clear_circular_list(gt_log->kfifo[gt_log->mod]);
	}

unlocked:
	gt_mutex_unlock(gt_log->log_mutex);

	return count;
}

const struct file_operations s_gt_dpu_log_fops = {
	.owner = THIS_MODULE,
	.open  = gtdpu_log_open,
	.read  = gtdpu_log_read,
	.write = gtdpu_log_write,
	.llseek  = no_llseek,
};
/* used by drm dpu */

GT_EXT_SYM(gt_dev_printk);
GT_EXT_SYM(gt_printk);
GT_EXT_SYM(gt_print_hex_dump);
GT_EXT_SYM(gtdpu_info);
GT_EXT_SYM(gtdpu_warn);
GT_EXT_SYM(gtdpu_err);
GT_EXT_SYM(gtdpu_log_init);
GT_EXT_SYM(gtdpu_log_fini);
GT_EXT_SYM(gt_seq_printf);
GT_EXT_SYM(s_gt_dpu_log_fops);
GT_EXT_SYM(s_dpu_debug);
