/*
* Copyright (c) Gengtu Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <linux/version.h>
#include <linux/kthread.h>
#include <linux/kernel.h>
#include <linux/workqueue.h>
#include <linux/slab.h>
#include <linux/freezer.h>
#include <linux/io.h>
#include <linux/pid.h>
#include <linux/sched.h>
#include <linux/interrupt.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)) */

#include "gt_misc.h"
#include "gt_task.h"

#define KERNEL_ID 0xffffffffL

struct gt_work {
	struct work_struct work;
	void (*func)(void *data);
	void *data;
};

struct gt_dwork {
	struct delayed_work dwork;
	void (*func)(void *data);
	void *data;
};

gt_task *gt_get_current_task(void)
{
	return current;
}
GT_EXT_SYM(gt_get_current_task);

void gt_enable_oom_killer(void)
{
	current->flags &= ~PF_DUMPCORE;
}
GT_EXT_SYM(gt_enable_oom_killer);

void gt_disable_oom_killer(void)
{
	WARN_ON(current->flags & PF_DUMPCORE);
	current->flags |= PF_DUMPCORE;
}
GT_EXT_SYM(gt_disable_oom_killer);

void gt_task_set_current_state(long state)
{
	set_current_state(TASK_INTERRUPTIBLE);
}
GT_EXT_SYM(gt_task_set_current_state);

void gt_task_schedule()
{
	schedule();
}
GT_EXT_SYM(gt_task_schedule);

void gt_task_set_user_nice(gt_task *task, long nice)
{
	set_user_nice((struct task_struct *)task, nice);
}
GT_EXT_SYM(gt_task_set_user_nice);

int gt_task_get_current_nice(void)
{
	return task_nice(current);
}
GT_EXT_SYM(gt_task_get_current_nice);

int gt_task_stop(gt_task *task)
{
	return kthread_stop((struct task_struct *)task);
}
GT_EXT_SYM(gt_task_stop);

void gt_tasklet_kill(gt_task *task)
{
	tasklet_kill((struct tasklet_struct *)task);
}
GT_EXT_SYM(gt_tasklet_kill);

void gt_do_exit(long code)
{
#ifdef GTGPU_DO_EXIT_PRESENT
	do_exit(code);
#endif
}
GT_EXT_SYM(gt_do_exit);

uint64_t gt_task_size(void)
{
	return TASK_SIZE;
}
GT_EXT_SYM(gt_task_size);

uint32_t gt_task_id(void)
{
	if (in_interrupt())
		return KERNEL_ID;

	return task_tgid_nr(current);
}
GT_EXT_SYM(gt_task_id);

uint32_t gt_task_pid(void)
{
	if (in_interrupt())
		return KERNEL_ID;

	return current->pid;
}
GT_EXT_SYM(gt_task_pid);

uint32_t gt_task_pid_vnr(void)
{
	if (in_interrupt())
		return KERNEL_ID;

	return task_pid_vnr(current);
}
GT_EXT_SYM(gt_task_pid_vnr);

uint32_t gt_task_virtual_id(void)
{
	if (in_interrupt())
		return KERNEL_ID;

	return task_tgid_vnr(current);
}
GT_EXT_SYM(gt_task_virtual_id);

gt_pid *gt_find_vpid(uint32_t nr)
{
	return find_vpid(nr);
}
GT_EXT_SYM(gt_find_vpid);

char *gt_get_task_name_by_pid(uint32_t nr)
{
	struct task_struct *task = pid_task(find_vpid(nr), PIDTYPE_PID);
	return task ? task->comm : NULL;
}
GT_EXT_SYM(gt_get_task_name_by_pid);

int gt_kill_pid(gt_pid *pid, int sig, int priv)
{
	return kill_pid((struct pid *)pid, sig, priv);
}
GT_EXT_SYM(gt_kill_pid);

char *gt_task_name(void)
{
	return current->comm;
}
GT_EXT_SYM(gt_task_name);

bool gt_task_is_in_kernel(void)
{
	return current->mm == NULL;
}
GT_EXT_SYM(gt_task_is_in_kernel);

#ifdef CONFIG_FAULT_INJECTION
void gt_task_set_current_make_fail(bool val)
{
	return current->make_it_fail = val;
}
GT_EXT_SYM(gt_task_set_current_make_fail);
#endif

bool gt_set_freezable(void)
{
	return set_freezable();
}
GT_EXT_SYM(gt_set_freezable);

bool gt_kthread_freezable_should_stop(void)
{
	return kthread_freezable_should_stop(NULL);
}
GT_EXT_SYM(gt_kthread_freezable_should_stop);

void gt_get_task_struct(gt_task *task)
{
	get_task_struct((struct task_struct *)task);
}
GT_EXT_SYM(gt_get_task_struct);

void gt_put_task_struct(gt_task *task)
{
	put_task_struct((struct task_struct *)task);
}
GT_EXT_SYM(gt_put_task_struct);

gt_task *gt_task_create(int (*func)(void *data), void *data, const char name[])
{
	return kthread_run(func, data, "%s", name);
}
GT_EXT_SYM(gt_task_create);

gt_workqueue *gt_create_freezable_wq(const char *name)
{
	return create_freezable_workqueue(name);
}
GT_EXT_SYM(gt_create_freezable_wq);

gt_workqueue *gt_create_workqueue(const char *name)
{
	return create_workqueue(name);
}
GT_EXT_SYM(gt_create_workqueue);

gt_workqueue *gt_create_singlethread_wq(const char *name)
{
	return create_singlethread_workqueue(name);
}
GT_EXT_SYM(gt_create_singlethread_wq);

gt_workqueue *gt_alloc_workqueue(const char *name, unsigned int flags, int max_active)
{
	return alloc_workqueue(name, flags, max_active);
}
GT_EXT_SYM(gt_alloc_workqueue);

void gt_destroy_workqueue(gt_workqueue *wq)
{
	destroy_workqueue((struct workqueue_struct *)wq);
}
GT_EXT_SYM(gt_destroy_workqueue);

static void gt_work_func(struct work_struct *data)
{
	struct gt_work *work = container_of(data, struct gt_work, work);
	work->func(work->data);
}

void *gt_work_alloc(void (*func)(void *data), void *data)
{
	struct gt_work *work = (struct gt_work *)
		kzalloc(sizeof(struct gt_work), GFP_KERNEL);
	if (!work)
		return NULL;

	work->func = func;
	work->data = data;
	INIT_WORK(&work->work, gt_work_func);
	return work;
}
GT_EXT_SYM(gt_work_alloc);

void gt_work_destroy(void *work)
{
	kfree(work);
}
GT_EXT_SYM(gt_work_destroy);

int gt_queue_work(gt_workqueue *wq, void *wk)
{
	return queue_work((struct workqueue_struct *)wq, &((struct gt_work *)wk)->work);
}
GT_EXT_SYM(gt_queue_work);

int gt_queue_work_on(int cpu, gt_workqueue *wq, void *wk)
{
	return queue_work_on(cpu, (struct workqueue_struct *)wq, &((struct gt_work *)wk)->work);
}
GT_EXT_SYM(gt_queue_work_on);

static struct delayed_work *gt_to_dwork(struct work_struct *work)
{
	return container_of(work, struct delayed_work, work);
}

static void gt_dwork_func(struct work_struct *work)
{
	struct delayed_work *dwork = gt_to_dwork(work);

	struct gt_dwork *dwk = container_of(dwork, struct gt_dwork, dwork);
	dwk->func(dwk->data);
}

void *gt_dwork_alloc(void (*func)(void *data), void *data)
{
	struct gt_dwork *dwork = (struct gt_dwork *)
		kzalloc(sizeof(struct gt_dwork), GFP_KERNEL);
	if (!dwork)
		return NULL;

	dwork->func = func;
	dwork->data = data;
	INIT_DELAYED_WORK(&dwork->dwork, gt_dwork_func);
	return dwork;
}
GT_EXT_SYM(gt_dwork_alloc);

int gt_queue_dwork(gt_workqueue *wq, gt_dwork *dwk, unsigned int msec)
{
	return queue_delayed_work((struct workqueue_struct *)wq,
		&((struct gt_dwork *)dwk)->dwork, msecs_to_jiffies(msec));
}
GT_EXT_SYM(gt_queue_dwork);

int gt_schedule_dwork(gt_dwork *dwk, unsigned int msec)
{
	return schedule_delayed_work(&((struct gt_dwork *)dwk)->dwork, msecs_to_jiffies(msec));
}
GT_EXT_SYM(gt_schedule_dwork);

int gt_mod_dwork(gt_dwork *dwk, unsigned int msec)
{
	return mod_delayed_work(system_wq, &((struct gt_dwork *)dwk)->dwork, msecs_to_jiffies(msec));
}
GT_EXT_SYM(gt_mod_dwork);

int gt_cancel_dwork(gt_dwork *dwk)
{
	return cancel_delayed_work(&((struct gt_dwork *)dwk)->dwork);
}
GT_EXT_SYM(gt_cancel_dwork);

int gt_cancel_dwork_sync(gt_dwork *dwk)
{
	return cancel_delayed_work_sync(&((struct gt_dwork *)dwk)->dwork);
}
GT_EXT_SYM(gt_cancel_dwork_sync);

void gt_dwork_destroy(void *dwork)
{
	kfree(dwork);
}
GT_EXT_SYM(gt_dwork_destroy);

bool gt_schedule_work(void *wk)
{
	return schedule_work(&((struct gt_work *)wk)->work);
}
GT_EXT_SYM(gt_schedule_work);

void gt_flush_scheduled_work(void)
{
#if defined(GTGPU_ATRRIBUTE_WARNING_PRESENT)
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0))
	flush_scheduled_work();
#else
	flush_workqueue(system_wq);
#endif
#endif
}
GT_EXT_SYM(gt_flush_scheduled_work);

bool gt_cancel_work_sync(void *wk)
{
	return cancel_work_sync(&((struct gt_work *)wk)->work);
}
GT_EXT_SYM(gt_cancel_work_sync);

void gt_flush_workqueue(gt_workqueue *wq)
{
	flush_workqueue((struct workqueue_struct *)wq);
}
GT_EXT_SYM(gt_flush_workqueue);

bool gt_flush_work(gt_work *work)
{
	return flush_work(&((struct gt_work *)work)->work);
}
GT_EXT_SYM(gt_flush_work);

signed long  gt_schedule_timeout(signed long timeout)
{
	return schedule_timeout(timeout);
}
GT_EXT_SYM(gt_schedule_timeout);

void gt_might_sleep(void)
{
	might_sleep();
}
GT_EXT_SYM(gt_might_sleep);

unsigned int gt_work_busy(gt_work *work)
{
	return work_busy(&((struct gt_work *)work)->work);
}
GT_EXT_SYM(gt_work_busy);
