/*
* Copyright (c) Gengtu Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <linux/version.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include "gt_drm_version.h"
#if (DRM_VERSION >= KERNEL_VERSION(4, 11, 0))
#include <linux/sched/clock.h>
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif /* (DRM_VERSION >= KERNEL_VERSION(4, 11, 0)) */

#include "gt_misc.h"
#include "gt_timer.h"
#include <linux/jiffies.h>
#include <linux/ktime.h>

struct gt_timer {
	struct timer_list timer;
	void (*function)(unsigned long data);
	unsigned long data;
};

unsigned long gt_time_jiffies(void)
{
	return jiffies;
}
GT_EXT_SYM(gt_time_jiffies);

bool gt_time_after(unsigned long timeout)
{
	return time_after(jiffies, timeout);
}
GT_EXT_SYM(gt_time_after);

unsigned long gt_msecs_to_jiffies(unsigned int m)
{
	return msecs_to_jiffies(m);
}
GT_EXT_SYM(gt_msecs_to_jiffies);

unsigned int gt_jiffies_to_msecs(unsigned long j)
{
	return jiffies_to_msecs(j);
}
GT_EXT_SYM(gt_jiffies_to_msecs);

unsigned long gt_usecs_to_jiffies(unsigned int u)
{
	return usecs_to_jiffies(u);
}
GT_EXT_SYM(gt_usecs_to_jiffies);

unsigned int gt_jiffies_to_usecs(unsigned long j)
{
	return jiffies_to_usecs(j);
}
GT_EXT_SYM(gt_jiffies_to_usecs);

bool gt_time_before(unsigned long timeout)
{
	return time_before(jiffies, timeout);
}
GT_EXT_SYM(gt_time_before);

struct gt_timer *gt_timer_alloc(void)
{
	struct gt_timer *gt_timer = (struct gt_timer *)
		kzalloc(sizeof(struct gt_timer), GFP_KERNEL);
	return gt_timer;
}
GT_EXT_SYM(gt_timer_alloc);

void gt_timer_free(struct gt_timer *gt_timer)
{
	kfree(gt_timer);
}
GT_EXT_SYM(gt_timer_free);

#if (DRM_VERSION >= KERNEL_VERSION(4, 15, 0))
static void gt_timer_callback_wrapper(struct timer_list *timer)
{
	struct gt_timer *gt_timer = container_of(timer, struct gt_timer, timer);
	gt_timer->function(gt_timer->data);
}
#else
static void gt_timer_callback_wrapper(unsigned long data)
{
	struct gt_timer *gt_timer = (struct gt_timer *)data;
	gt_timer->function(gt_timer->data);
}
#endif

void gt_timer_setup(struct gt_timer *gt_timer,
                      void (*function)(unsigned long data), unsigned long data)
{
	gt_timer->function = function;
	gt_timer->data = data;

#if (DRM_VERSION >= KERNEL_VERSION(4, 15, 0))
	timer_setup(&gt_timer->timer, gt_timer_callback_wrapper, 0);
#else
	init_timer(&gt_timer->timer);
	gt_timer->timer.function = gt_timer_callback_wrapper;
	gt_timer->timer.data = (unsigned long)gt_timer;
#endif
}
GT_EXT_SYM(gt_timer_setup);

void gt_timer_start(struct gt_timer *gt_timer, unsigned int delay)
{
	gt_timer->timer.expires = delay + jiffies;
	add_timer(&gt_timer->timer);
}
GT_EXT_SYM(gt_timer_start);

void gt_timer_stop(struct gt_timer *gt_timer)
{
	del_timer_sync(&gt_timer->timer);
}
GT_EXT_SYM(gt_timer_stop);

uint64_t gt_clockns64(void)
{
	uint64_t timenow;
	preempt_disable();
	timenow = sched_clock();
	preempt_enable();
	return timenow;
}
GT_EXT_SYM(gt_clockns64);

uint64_t gt_clockmonotonic(void)
{
	ktime_t time = ktime_get();

#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return time;
#else
	return time.tv64;
#endif
}
GT_EXT_SYM(gt_clockmonotonic);

uint64_t gt_clockmonotonic_raw(void)
{
#if (DRM_VERSION >= KERNEL_VERSION(5, 6, 0))
	struct timespec64 ts;
	ktime_get_raw_ts64(&ts);
	return ts.tv_sec * 1000000000 + ts.tv_nsec;
#else
	struct timespec ts;
	getrawmonotonic(&ts);
	return (uint64_t) ts.tv_sec * 1000000000 + ts.tv_nsec;
#endif
}
GT_EXT_SYM(gt_clockmonotonic_raw);

void gt_ndelay(uint64_t ns)
{
	ndelay(ns);
}
GT_EXT_SYM(gt_ndelay);

void gt_udelay(uint64_t us)
{
	udelay(us);
}
GT_EXT_SYM(gt_udelay);

void gt_mdelay(unsigned int ms)
{
	mdelay(ms);
}
GT_EXT_SYM(gt_mdelay);

void gt_msleep(uint32_t ms)
{
	msleep(ms);
}
GT_EXT_SYM(gt_msleep);

void gt_usleep_range(unsigned long min, unsigned long max)
{
	usleep_range(min, max);
}
GT_EXT_SYM(gt_usleep_range);

gt_ktime gt_ktime_get(void)
{
#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return ktime_get();
#else
	ktime_t ktime = {0};
	ktime = ktime_get();
	return ktime.tv64;
#endif
}
GT_EXT_SYM(gt_ktime_get);

gt_ktime gt_ktime_sub(gt_ktime end, gt_ktime start)
{
#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return ktime_sub(end, start);
#else
	ktime_t ktime = {0};
	ktime_t gt_start = {0};
	ktime_t gt_end = {0};
	gt_start.tv64 = start;
	gt_end.tv64 = end;
	ktime = ktime_sub(gt_end, gt_start);
	return ktime.tv64;
#endif
}
GT_EXT_SYM(gt_ktime_sub);

int gt_ktime_compare(const gt_ktime cmp1, const gt_ktime cmp2)
{
#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return ktime_compare(cmp1, cmp2);
#else
	ktime_t gt_cmp1 = {0};
	ktime_t gt_cmp2 = {0};
	gt_cmp1.tv64 = cmp1;
	gt_cmp2.tv64 = cmp2;
	return ktime_compare(gt_cmp1, gt_cmp2);
#endif
}
GT_EXT_SYM(gt_ktime_compare);

gt_ktime gt_ktime_add_us(const gt_ktime kt, const u64 usec)
{
#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return ktime_add_us(kt, usec);
#else
	ktime_t ktime = {0};
	ktime_t gt_kt = {0};
	gt_kt.tv64 = kt;
	ktime = ktime_add_us(gt_kt, usec);
	return ktime.tv64;
#endif
}
GT_EXT_SYM(gt_ktime_add_us);

s64 gt_ktime_to_us(const gt_ktime kt)
{
#if (DRM_VERSION >= KERNEL_VERSION(4, 10, 0))
	return ktime_to_us(kt);
#else
	ktime_t gt_kt = {0};
	gt_kt.tv64 = kt;
	return ktime_to_us(gt_kt);
#endif
}
GT_EXT_SYM(gt_ktime_to_us);
