/*
* Copyright (c) Gengtu Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/debugfs.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include "gt_srvkm.h"
#include "gt_misc.h"
#include "gt_debugfs.h"
#include "gt69_debugfs.h"
#include "gt69_procfs.h"

static int gt_debugfs_file_open(struct inode *inode, struct file *file)
{
	return gt_debugfs_open(inode, file, inode->i_private);
}

static int gt_debugfs_file_close(struct inode *inode, struct file *file)
{
	return gt_debugfs_close(inode, file, inode->i_private);
}

static ssize_t gt_debugfs_file_read(struct file *file, char __user *buf,
		size_t count, loff_t *pos)
{
	return gt_debugfs_read(file, buf, count, pos);
}

static loff_t gt_debugfs_file_lseek(struct file *file, loff_t off, int origin)
{
	return gt_debugfs_lseek(file, off, origin);
}

static ssize_t gt_debugfs_file_write(struct file *file, const char __user *buffer,
		size_t count, loff_t *pos)
{
	return gt_debugfs_write(file, buffer, count, pos);
}

static const struct file_operations gt_debugfs_file_gen_ops = {
	.owner   = THIS_MODULE,
	.open    = gt_debugfs_file_open,
	.release = gt_debugfs_file_close,
	.read    = gt_debugfs_file_read,
	.llseek  = gt_debugfs_file_lseek,
	.write   = gt_debugfs_file_write,
};

static const struct file_operations gt_debugfs_file_rnd_ops = {
	.owner  = THIS_MODULE,
	.read   = gt_debugfs_file_read,
	.llseek = gt_debugfs_file_lseek,
	.write  = gt_debugfs_file_write,
};

static void *gt_seq_start_ops(struct seq_file *seqf, loff_t *pos)
{
	void *entry = seqf->private;
	return gt_debugfs_seq_start(entry, pos);
}

static void gt_seq_stop_ops(struct seq_file *seqf, void *priv)
{
	void *entry = seqf->private;
	gt_debugfs_seq_stop(entry, priv);
}

static void *gt_seq_next_ops(struct seq_file *seqf, void *priv, loff_t *pos)
{
	void *entry = seqf->private;
	return gt_debugfs_seq_next(entry, priv, pos);
}

static int gt_seq_show_ops(struct seq_file *seqf, void *priv)
{
	void *entry = seqf->private;
	return gt_debugfs_seq_show(entry, priv);
}

static struct seq_operations gt_seq_ops = {
	.start = gt_seq_start_ops,
	.stop  = gt_seq_stop_ops,
	.next  = gt_seq_next_ops,
	.show  = gt_seq_show_ops,
};

int gt_seq_open_for_gt69(gt_file *file)
{
	return seq_open((struct file *)file, &gt_seq_ops);
}
GT_EXT_SYM(gt_seq_open_for_gt69);

int gt_single_open_for_gt69(gt_file *file)
{
	return single_open((struct file *)file, gt_seq_show_ops, NULL);
}
GT_EXT_SYM(gt_single_open_for_gt69);

void gt_seq_set_prvdata(gt_seq_file *seqf, void *priv)
{
	((struct seq_file *)seqf)->private = priv;
}
GT_EXT_SYM(gt_seq_set_prvdata);

void *gt_seq_get_prvdata(gt_seq_file *seqf)
{
	return ((struct seq_file *)seqf)->private;
}
GT_EXT_SYM(gt_seq_get_prvdata);

int gt_seq_release(gt_inode *inode, gt_file *file)
{
	return seq_release((struct inode *)inode, (struct file *)file);
}
GT_EXT_SYM(gt_seq_release);

int gt_single_release(gt_inode *inode, gt_file *file)
{
	return single_release((struct inode *)inode, (struct file *)file);
}
GT_EXT_SYM(gt_single_release);

void *gt_debugfs_file_get_prvdata(gt_file *file)
{
	return ((struct file *)file)->f_path.dentry->d_inode->i_private;
}
GT_EXT_SYM(gt_debugfs_file_get_prvdata);

const char *gt_debugfs_file_name(gt_file *file)
{
	return ((struct file *)file)->f_path.dentry->d_name.name;
}
GT_EXT_SYM(gt_debugfs_file_name);

ssize_t gt_seq_read(gt_file *file, char __user *buf, size_t count, loff_t *pos)
{
	return seq_read((struct file *)file, buf, count, pos);
}
GT_EXT_SYM(gt_seq_read);

int gt_seq_write(gt_seq_file *seq, const void *data, size_t len)
{
	return seq_write(seq, data, len);
}
GT_EXT_SYM(gt_seq_write);

loff_t gt_seq_lseek(gt_file *file, loff_t off, int origin)
{
	return seq_lseek((struct file *)file, off, origin);
}
GT_EXT_SYM(gt_seq_lseek);

loff_t gt_file_pos(gt_file *file)
{
	return ((struct file *)file)->f_pos;
}
GT_EXT_SYM(gt_file_pos);

void gt_file_set_pos(gt_file *file, loff_t pos)
{
	((struct file *)file)->f_pos = pos;
}
GT_EXT_SYM(gt_file_set_pos);

void gt_seq_vprintf(gt_seq_file *seqf, const char *fmt, va_list args)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0)
	seq_vprintf(seqf, fmt, args);
#else
	char buf[512];
	vsnprintf(buf, 512, fmt, args);
	seq_printf(seqf, "%s", buf);
#endif
}
GT_EXT_SYM(gt_seq_vprintf);

void gt_seq_puts(gt_seq_file *seqf, const char *s)
{
	seq_puts(seqf, s);
}
GT_EXT_SYM(gt_seq_puts);

bool gt_seq_has_overflowed(gt_seq_file *seqf)
{
	struct seq_file *file = (struct seq_file *)seqf;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	return seq_has_overflowed(file);
#else
	return file->count == file->size;
#endif
}
GT_EXT_SYM(gt_seq_has_overflowed);

gt_dentry *gt_debugfs_create_file_for_gt69(const char *name, umode_t mode,
		gt_dentry *parent, void *data, bool is_gen)
{
	if (is_gen)
		return debugfs_create_file(name, mode, parent, data, &gt_debugfs_file_gen_ops);
	else
		return debugfs_create_file(name, mode, parent, data, &gt_debugfs_file_rnd_ops);
}
GT_EXT_SYM(gt_debugfs_create_file_for_gt69);

void gt_debugfs_remove(gt_dentry *dentry)
{
	debugfs_remove((struct dentry *)dentry);
}
GT_EXT_SYM(gt_debugfs_remove);

gt_dentry *gt_debugfs_create_dir(const char *name, gt_dentry *parent)
{
	return debugfs_create_dir(name, parent);
}
GT_EXT_SYM(gt_debugfs_create_dir);

int gt_debugorprocfs_register(void)
{
#if defined(ANDROID) && !defined(__GT_CONTAINER__)
#if defined(CONFIG_PROC_FS)
	return PVRProcFsRegister();
#elif defined(CONFIG_DEBUG_FS)
	return PVRDebugFsRegister();
#endif /* defined(CONFIG_PROC_FS) || defined(CONFIG_DEBUG_FS) */
#else
#if defined(CONFIG_DEBUG_FS)
	return PVRDebugFsRegister();
#elif defined(CONFIG_PROC_FS)
	return PVRProcFsRegister();
#endif /* defined(CONFIG_DEBUG_FS) || defined(CONFIG_PROC_FS) */
#endif /* defined(ANDROID) */
}
GT_EXT_SYM(gt_debugorprocfs_register);

GT_EXT_SYM(gt_debugfs_seq_start);
GT_EXT_SYM(gt_debugfs_seq_stop);
GT_EXT_SYM(gt_debugfs_seq_next);
GT_EXT_SYM(gt_debugfs_seq_show);
GT_EXT_SYM(gt_debugfs_open);
GT_EXT_SYM(gt_debugfs_close);
GT_EXT_SYM(gt_debugfs_read);
GT_EXT_SYM(gt_debugfs_lseek);
GT_EXT_SYM(gt_debugfs_write);
